#include <stdio.h>

/*
Copyright (c) 2021 Devine Lu Linvega

Permission to use, copy, modify, and distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE.
*/

#define SIZE 2

typedef unsigned char Uint8;
typedef unsigned short Uint16;

typedef struct {
	char raw[8 * SIZE][8 * SIZE + 1];
	Uint8 width, height, data[4][8], pad;
} Glyph;

Uint8 maxwidth = 0, maxheight = 0;
Glyph glyphs[256];

Uint8 spritesheet[256 + 256 * SIZE * SIZE * 8];

static char *
scpy(char *src, char *dst, int len)
{
	int i = 0;
	while((dst[i] = src[i]) && i < len - 2) i++;
	dst[i + 1] = '\0';
	return dst;
}

static int
slen(char *s)
{
	int i = 0;
	while(s[i] && s[++i])
		;
	return i;
}

static Uint8
chex(char c)
{
	if(c >= 'a' && c <= 'f')
		return 10 + c - 'a';
	if(c >= 'A' && c <= 'F')
		return 10 + c - 'A';
	return (c - '0') & 0xF;
}

static Uint16
shex(char *s, int len)
{
	int i, n = 0;
	for(i = 0; i < len; ++i)
		n |= (chex(s[i]) << ((len - i - 1) * 4));
	return n;
}

static int
isdata(char *line)
{
	int i = 0;
	for(i = 0; i < slen(line); ++i) {
		if(line[i] != '-' && line[i] != '.' && line[i] != '@')
			return 0;
	}
	return 1;
}

static void
cleanup(Glyph *g)
{
	int x, y;
	for(y = 0; y < 8 * SIZE; ++y) {
		for(x = 0; x < 8 * SIZE; ++x) {
			if(g->raw[y][x] == '@')
				g->raw[y][x] = '@';
			else
				g->raw[y][x] = '.';
		}
		g->raw[y][8 * SIZE] = 0;
	}
}

static Uint8
getbyte(Glyph *g, Uint8 x, Uint8 y)
{
	Uint8 num = 0;
	int p;
	if(!g->width)
		return 0;
	for(p = 0; p < 8; ++p)
		if(g->raw[y][x + 7 - p] == '@')
			num |= (1 << p);
	return num;
}

static int
error(char *name, char *id)
{
	fprintf(stderr, "Error: %s[%s]\n", name, id);
	return 0;
}

static int
pass1(FILE *f)
{
	Uint8 id = 0;
	char w[64];
	while(fscanf(f, "%s", w) == 1) {
		if(w[0] == '0' && w[1] == 'x' && w[4] == ':') {
			id = shex(w + 2, 2);
			glyphs[id].height = 0;
			continue;
		} else if(!isdata(w))
			continue;
		if(glyphs[id].width && slen(w) != glyphs[id].width)
			return error("Changing width", w);
		scpy(w, glyphs[id].raw[glyphs[id].height], slen(w) + 1);
		glyphs[id].width = slen(w);
		if(glyphs[id].height > SIZE * 8) {
			printf("[%02x] glyph size: %d/%d\n", id, glyphs[id].height, SIZE * 8);
			return error("Reached max height", "");
		}
		if(glyphs[id].width > SIZE * 8) {
			printf("%d(%02x)", glyphs[id].width, id);
			return error("Reached max width", "");
		}
		glyphs[id].height++;
	}
	return 1;
}

static int
pass2(void)
{
	int i, x, y, b;
	/* generate width table */
	for(i = 0; i < 256; ++i) {
		Glyph *g = &glyphs[i];
		spritesheet[i] = g->width;
		cleanup(g);
	}
	for(i = 0; i < 256; ++i) {
		Glyph *g = &glyphs[i];
		if(!g->width)
			continue;
		for(y = 0; y < SIZE; ++y) {
			for(x = 0; x < SIZE; ++x) {
				for(b = 0; b < 8; ++b) {
					int index = 256 + (i * SIZE * SIZE * 8);
					int byteid = index + y * 8 * SIZE + x * 8 + b;
					spritesheet[byteid] = getbyte(g, x * 8, y * 8 + b);
				}
			}
		}
	}
	return 1;
}

int
main(int argc, char *argv[])
{
	FILE *f;
	printf("> %s\n", argv[2]);
	if(argc < 2 || !(f = fopen(argv[1], "r")))
		return !error("Open", "Failed");
	if(!pass1(f) || !pass2())
		return !error("Assembly", "Failed");
	fwrite(spritesheet, 256 + 256 * SIZE * SIZE * 8, 1, fopen(argv[2], "wb"));
	return 0;
}
